/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/psp/framebuffer.c: Framebuffer management functionality for the
 * PSP.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/psp/framebuffer.h"
#include "src/sysdep/psp/ge-util.h"
#include "src/sysdep/psp/internal.h"

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int sys_framebuffer_supported(void)
{
    return 1;
}

/*-----------------------------------------------------------------------*/

SysFramebuffer *sys_framebuffer_create(int width, int height)
{
    SysFramebuffer *framebuffer = mem_alloc(sizeof(*framebuffer), 0, 0);
    if (UNLIKELY(!framebuffer)) {
        DLOG("Failed to allocate SysFramebuffer");
        return NULL;
    }

    framebuffer->width  = width;
    framebuffer->height = height;
    framebuffer->stride = align_up(framebuffer->width, 4);
    const uint32_t size = framebuffer->stride * framebuffer->height * 4;
    framebuffer->pixels = psp_vram_alloc(size);
    if (UNLIKELY(!framebuffer->pixels)) {
        DLOG("Failed to allocate VRAM for %dx%d framebuffer (%u bytes)",
             width, height, size);
        mem_free(framebuffer);
        return NULL;
    }

    framebuffer->texture.width     = framebuffer->width;
    framebuffer->texture.height    = framebuffer->height;
    framebuffer->texture.stride    = framebuffer->stride;
    framebuffer->texture.format    = GE_TEXFMT_8888;
    framebuffer->texture.swizzled  = 0;
    framebuffer->texture.mipmaps   = 0;
    framebuffer->texture.pixels[0] = framebuffer->pixels;
    framebuffer->texture.palette   = NULL;
    framebuffer->texture.repeat_u  = 0;
    framebuffer->texture.repeat_v  = 0;
    framebuffer->texture.antialias = 1;
    framebuffer->texture.lock_buf  = NULL;

    return framebuffer;
}

/*-----------------------------------------------------------------------*/

void sys_framebuffer_destroy(SysFramebuffer *framebuffer)
{
    if (psp_current_framebuffer() == framebuffer) {
        psp_use_framebuffer(NULL);
    }
    if (psp_current_texture() == &framebuffer->texture) {
        sys_texture_apply(0, NULL);
    }

    /* Make sure we sync before freeing the VRAM, so the GE doesn't stomp
     * on subsequently-allocated buffers. */
    ge_sync();

    psp_vram_free(framebuffer->pixels);
    mem_free(framebuffer);
}

/*-----------------------------------------------------------------------*/

void sys_framebuffer_bind(SysFramebuffer *framebuffer)
{
    if (framebuffer) {
        psp_use_framebuffer(framebuffer);
    } else {
        psp_use_framebuffer(NULL);
    }
}

/*-----------------------------------------------------------------------*/

SysTexture *sys_framebuffer_get_texture(SysFramebuffer *framebuffer)
{
    return &framebuffer->texture;
}

/*-----------------------------------------------------------------------*/

void sys_framebuffer_set_antialias(SysFramebuffer *framebuffer, int on)
{
    sys_texture_set_antialias(&framebuffer->texture, on);
}

/*-----------------------------------------------------------------------*/

void sys_framebuffer_discard_data(UNUSED SysFramebuffer *framebuffer)
{
    /* Ignore (we can't make use of the hint). */
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
